@@ -38,6 +38,7 @@ var filetypeMime = map[string]string{
 	"htm":  "text/html",
 	"html": "text/html",
 	"js":   "application/javascript",
+	"mjs":  "application/javascript",
 	"json": "application/json",
 	"svg":  "image/svg+xml",
 	"xml":  "text/xml",
@@ -389,7 +390,7 @@ func run() int {
 	m.Add("text/css", cssMinifier)
 	m.Add("text/html", htmlMinifier)
 	m.Add("image/svg+xml", svgMinifier)
-	m.AddRegexp(regexp.MustCompile("^(application|text)/(x-)?(java|ecma)script$"), jsMinifier)
+	m.AddRegexp(regexp.MustCompile("^(application|text)/(x-)?(java|ecma|j|live)script(1\\.[0-5])?$|^module$"), jsMinifier)
 	m.AddRegexp(regexp.MustCompile("[/+]json$"), jsonMinifier)
 	m.AddRegexp(regexp.MustCompile("[/+]xml$"), xmlMinifier)
 
