@@ -55,7 +55,7 @@ func TestDataURI(t *testing.T) {
 	m.AddFunc("text/x", func(_ *M, w io.Writer, r io.Reader, _ map[string]string) error {
 		b, _ := ioutil.ReadAll(r)
 		test.String(t, string(b), "<?xx?>")
-		w.Write(b)
+		_, _ = w.Write(b)
 		return nil
 	})
 	for _, tt := range dataURITests {
