@@ -36,6 +36,15 @@ func TestHexadecimalNumber(t *testing.T) {
 	test.Bytes(t, hexadecimalNumber([]byte("0x10000000000"), 0), []byte("0x10000000000"))
 }
 
+func TestString(t *testing.T) {
+	test.Bytes(t, minifyString([]byte(`""`), true), []byte(`""`))
+	test.Bytes(t, minifyString([]byte(`"abc"`), true), []byte(`"abc"`))
+	test.Bytes(t, minifyString([]byte(`'abc'`), true), []byte(`"abc"`))
+	test.Bytes(t, minifyString([]byte(`"\8\9\t"`), true), []byte("\"89\t\""))
+	test.Bytes(t, minifyString([]byte(`"\12"`), true), []byte(`"\n"`))
+	test.Bytes(t, minifyString([]byte(`"\n\r$"`), true), []byte("`\n\r\\$`"))
+}
+
 func TestHasSideEffects(t *testing.T) {
 	jsTests := []struct {
 		js  string
